@@ -60,7 +60,18 @@ typedef enum {
 	AUTORESTART_MODE_REQUESTS,
 } auto_restart_mode;
 
+#define URI_LIST_ITEM_SIZE 256
+#define ACTIVE_URI_LIST_ITEM_COUNT 100
+#define WAITING_URI_LIST_ITEM_COUNT 100
+
 typedef struct {
+	int32_t id;
+	time_t start_time;
+	char uri[URI_LIST_ITEM_SIZE];
+} uri_item;
+
+typedef struct {
+	uint32_t requests_counter;
 	uint32_t handled_requests;
 	time_t start_time;
 	char restart_issued;
@@ -68,6 +79,8 @@ typedef struct {
 	int active_requests;
 	int waiting_requests;
 	int accepting_requests;
+	uri_item active_uri_list[ACTIVE_URI_LIST_ITEM_COUNT];
+	uri_item waiting_uri_list[WAITING_URI_LIST_ITEM_COUNT];
 } dashboard_data;
 
 #define INITIAL_DATA_MAX_ALLOCA_SIZE 8192
@@ -105,6 +118,7 @@ typedef struct xsp_data {
 	char *debug;
 	char *env_vars;
 	char *iomap;
+	char *hidden;
 	char status; /* One of the FORK_* in the enum above.
 		      * Don't care if run_xsp is "false" */
 	char is_virtual; /* is the server virtual? */
@@ -371,6 +385,17 @@ apache_get_username ()
 }
 
 #ifndef APACHE13
+inline static void
+initialize_uri_list (uri_item* list, int nitems)
+{
+	int i;
+	for (i = 0; i < nitems; i++) {
+		memset (&list [i], 0, sizeof (uri_item));
+		list [i].id = -1;
+		list [i].start_time = -1;
+	}
+}
+
 static void
 ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 {
@@ -474,12 +499,16 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
           
 				xsp->dashboard = apr_shm_baseaddr_get (xsp->dashboard_shm);
 				xsp->dashboard->start_time = time (NULL);
+				xsp->dashboard->requests_counter = 0;
 				xsp->dashboard->handled_requests = 0;
 				xsp->dashboard->restart_issued = 0;
 				xsp->dashboard->active_requests = 0;
 				xsp->dashboard->waiting_requests = 0;
 				xsp->dashboard->starting = 0;
 				xsp->dashboard->accepting_requests = 1;
+
+				initialize_uri_list (xsp->dashboard->active_uri_list, ACTIVE_URI_LIST_ITEM_COUNT);
+				initialize_uri_list (xsp->dashboard->waiting_uri_list, WAITING_URI_LIST_ITEM_COUNT);
 			}
 		}
 	}
@@ -1644,6 +1673,13 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 				      "setgid: unable to set group id to %u. %s",
 				      (unsigned)apache_get_groupid (), strerror (errno));
 
+		DEBUG_PRINT (2, "initializing groups for forked process user %s", apache_get_username ());
+		if (initgroups (apache_get_username (), apache_get_groupid ()) == -1)
+			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
+				      "initgroups: unable to initialize supplementary group list for user %s: %s",
+				      apache_get_username (),
+				      strerror (errno));
+
 		DEBUG_PRINT (2, "switching forked process user to %s", apache_get_username ());
 		if (setuid (apache_get_userid ()) == -1)
 			ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
@@ -1737,6 +1773,11 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 		argv [argi++] = config->applications;
 	}
 
+	if (config->hidden != NULL) {
+		if (strcasecmp (config->hidden, "false"))
+			argv [argi++] = "--no-hidden";
+	}
+	
 	argv [argi++] = "--nonstop";
 	if (config->document_root != NULL) {
 		argv [argi++] = "--root";
@@ -2052,8 +2093,54 @@ send_initial_data (request_rec *r, apr_socket_t *sock, char auto_app)
 	return 0;
 }
 
+inline static void clear_uri_item (uri_item* list, int nitems, int32_t id)
+{
+	int i;
+
+	for (i = 0; i < nitems; i++) {
+		if (list [i].id != id)
+			continue;
+		list [i].id = -1;
+		list [i].start_time = -1;
+		return;
+	}
+}
+
+inline static void set_uri_item (uri_item* list, int nitems, request_rec* r, int32_t id)
+{
+	int i;
+	int uri_len;
+	int args_len;
+
+	for (i = 0; i < nitems; i++) {
+		if (list [i].id != -1)
+			continue;
+		list [i].id = id;
+		list [i].start_time = id == -1 ? -1 : time (NULL);
+		if (r->uri) {
+			uri_len = strlen (r->uri);
+			if (uri_len > URI_LIST_ITEM_SIZE)
+				uri_len = URI_LIST_ITEM_SIZE;
+			memcpy (list [i].uri, r->uri, uri_len);
+			list [i].uri [uri_len] = '\0';
+		}
+
+		if (r->args) {
+			args_len = strlen (r->args);
+			if (args_len + uri_len + 1 > URI_LIST_ITEM_SIZE)
+				args_len = URI_LIST_ITEM_SIZE - uri_len - 1;
+			if (args_len > 0) {
+				list [i].uri [uri_len] = '?';
+				memcpy (&list [i].uri [uri_len + 1], r->args, args_len);
+				list [i].uri [args_len + uri_len + 1] = '\0';
+			}
+		}
+		break;
+	}
+}
+
 static int
-increment_active_requests (xsp_data *conf)
+increment_active_requests (xsp_data *conf, request_rec *r, int32_t id)
 {
 #ifndef APACHE13
 	/* This function tries to increment the counters that limit
@@ -2105,6 +2192,7 @@ increment_active_requests (xsp_data *conf)
 		      conf->dashboard_lock_file, max_active_requests);
 		      
 		conf->dashboard->waiting_requests++;
+		set_uri_item (conf->dashboard->waiting_uri_list, WAITING_URI_LIST_ITEM_COUNT, r, id);
 
 		int retries = 20;
 		while (retries-- > 0) {
@@ -2120,9 +2208,21 @@ increment_active_requests (xsp_data *conf)
 		 		break;
 		}
 
+		{
+			int i;
+			for (i = 0; i < WAITING_URI_LIST_ITEM_COUNT; i++) {
+				if (conf->dashboard->waiting_uri_list[i].id == id) {
+					conf->dashboard->waiting_uri_list[i].id = -1;
+					conf->dashboard->waiting_uri_list[i].start_time = -1;
+					break;
+				}
+			}
+		}
+
 		// Hopefully we haven't lost the lock, but if we have we still
 		// want to at least attempt to decrement the counter.
 		conf->dashboard->waiting_requests--;
+		clear_uri_item (conf->dashboard->waiting_uri_list, WAITING_URI_LIST_ITEM_COUNT, id);
 
 	 	// If we got to the end of the loop and still too
 	 	// many requests are going, stop processing the
@@ -2142,7 +2242,8 @@ increment_active_requests (xsp_data *conf)
 		return 0;
 
 	conf->dashboard->active_requests++;
-	
+	set_uri_item (conf->dashboard->active_uri_list, ACTIVE_URI_LIST_ITEM_COUNT, r, id);
+
 	return 1;
 #else
 	return 1;
@@ -2150,7 +2251,7 @@ increment_active_requests (xsp_data *conf)
 }
 
  static void
-decrement_active_requests (xsp_data *conf)
+decrement_active_requests (xsp_data *conf, uint32_t id)
 {
 #ifndef APACHE13
 	apr_status_t rv;
@@ -2165,7 +2266,8 @@ decrement_active_requests (xsp_data *conf)
 	// get a lock, we had better attempt to decrement it.
 		
 	conf->dashboard->active_requests--;
-
+	clear_uri_item (conf->dashboard->active_uri_list, ACTIVE_URI_LIST_ITEM_COUNT, id);
+	
 	if (rv == APR_SUCCESS)
 		apr_global_mutex_unlock (conf->dashboard_mutex);
 #endif
@@ -2192,6 +2294,7 @@ mono_execute_request (request_rec *r, char auto_app)
 	char *socket_name = NULL;
 	int retrying, was_starting;
 	int is_global = 0;
+	int32_t id = -1;
 
 	config = ap_get_module_config (r->server->module_config, &mono_module);
 	DEBUG_PRINT (2, "config = 0x%p", config);
@@ -2263,8 +2366,10 @@ mono_execute_request (request_rec *r, char auto_app)
 	if (conf->dashboard_mutex && conf->dashboard) {
 		if (apr_global_mutex_lock (conf->dashboard_mutex) == APR_SUCCESS) {
 			int ok = conf->dashboard->accepting_requests;
-			if (ok)
-				ok = increment_active_requests (conf);
+			if (ok) {
+				id = conf->dashboard->requests_counter++;
+				ok = increment_active_requests (conf, r, id);
+			}
 			
 			apr_global_mutex_unlock (conf->dashboard_mutex);
 
@@ -2290,7 +2395,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		// Note that rv's value after the loop ends is important.
 		if (rv != APR_SUCCESS) {
 			if (rv != -1) {
-				decrement_active_requests (conf);
+				decrement_active_requests (conf, id);
 				return HTTP_SERVICE_UNAVAILABLE;
 			}
 			DEBUG_PRINT (2, "No backend found, will start a new copy.");
@@ -2306,7 +2411,7 @@ mono_execute_request (request_rec *r, char auto_app)
 				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv2),
 					      "Failed to acquire %s lock, cannot continue starting new process",
 					      conf->dashboard_lock_file);
-				decrement_active_requests (conf);
+				decrement_active_requests (conf, id);
 				return HTTP_SERVICE_UNAVAILABLE;
 			}
 
@@ -2322,7 +2427,7 @@ mono_execute_request (request_rec *r, char auto_app)
 				if (retrying < 0) {
 					ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
 						      "Another process is attempting to start the backend. This request will fail.");
-					decrement_active_requests (conf);
+					decrement_active_requests (conf, id);
 					return HTTP_SERVICE_UNAVAILABLE;
 				}
 				connect_attempts++; /* keep trying */
@@ -2362,7 +2467,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		/* Failed to connect to mod-mono-server after several attempts. */
 		ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
 			      "Failed to connect to mod-mono-server after several attempts to spawn the process.");
-		decrement_active_requests (conf);
+		decrement_active_requests (conf, id);
 		return HTTP_SERVICE_UNAVAILABLE;
 	}
   
@@ -2371,7 +2476,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 			      "Failed to send initial data. %s", strerror (errno));
 		apr_socket_close (sock);
-		decrement_active_requests (conf);
+		decrement_active_requests (conf, id);
 		return HTTP_SERVICE_UNAVAILABLE;
 	}
 	
@@ -2391,7 +2496,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		status = HTTP_INTERNAL_SERVER_ERROR;
 	}
 	
-	decrement_active_requests (conf);
+	decrement_active_requests (conf, id);
 
 #ifndef APACHE13
 	if (conf->restart_mode > AUTORESTART_MODE_NONE) {
@@ -2744,6 +2849,22 @@ set_accepting_requests (void *data, char *alias, int accepting_requests)
 	}
 }
 
+inline static void
+send_uri_list (uri_item* list, int nitems, request_rec *r)
+{
+	int i;
+	char *buffer;
+
+	request_send_response_string (r, "<dl>\n");
+	for (i = 0; i < nitems; i++) {
+		if (list [i].id != -1) {
+			buffer = apr_psprintf (r->pool, "<dd>%d %ds %s</dd>\n", list [i].id, (int)(time (NULL) - list [i].start_time), list [i].uri);
+			request_send_response_string (r, buffer);
+		}
+	}
+	request_send_response_string (r, "</dl></li>");
+}
+
 static int
 mono_control_panel_handler (request_rec *r)
 {
@@ -2810,14 +2931,18 @@ mono_control_panel_handler (request_rec *r)
 					request_send_response_string(r, buffer);
 				}
 
-				buffer = apr_psprintf (r->pool, "<li>%d requests currently being processed; limit: %s</li>\n",
-						       xsp->dashboard->active_requests, xsp->max_active_requests);
+				buffer = apr_psprintf (r->pool, "<li>%d requests currently being processed; limit: %s; total: %d\n",
+						       xsp->dashboard->active_requests,
+						       xsp->max_active_requests ? xsp->max_active_requests : "unlimited",
+						       xsp->dashboard->requests_counter);
 				request_send_response_string(r, buffer);
-				
-				buffer = apr_psprintf (r->pool, "<li>%d requests currently waiting to be processed; limit: %s</li>\n",
-						       xsp->dashboard->waiting_requests, xsp->max_waiting_requests);
+
+
+				buffer = apr_psprintf (r->pool, "<li>%d requests currently waiting to be processed; limit: %s\n",
+						       xsp->dashboard->waiting_requests,
+						       xsp->max_waiting_requests ? xsp->max_active_requests : "unlimited");
 				request_send_response_string(r, buffer);
-				
+				send_uri_list (xsp->dashboard->waiting_uri_list, WAITING_URI_LIST_ITEM_COUNT, r);
 
 				rv = apr_global_mutex_unlock (xsp->dashboard_mutex);
 				if (rv != APR_SUCCESS)
@@ -3131,6 +3256,14 @@ static const command_rec mono_cmds [] = {
 		    "with the maximum number of requests specified by MonoMaxActiveRequests. "
 		    "Requests that can't be processed or held are dropped with Service Unavailable." 
 		    "Default value: 20"),
+	MAKE_CMD12 (MonoCheckHiddenFiles, hidden,
+		    "Do not protect hidden files/directories from being accessed by clients. Hidden files/directories are those with "
+		    "Hidden attribute on Windows and whose name starts with a dot on Unix. Any file/directory below a hidden directory "
+		    "is inacessible. This option turns the default behavior of protecting such locations off. If your application "
+		    "does not contain any hidden files/directories, you might want to use this option as the checking process has a "
+		    "per-request cost. Accepts a boolean value - 'true' or 'false'"
+		    "Default value: true. "
+		    "AppSettings key name: MonoServerCheckHiddenFiles. "),
 	{ NULL }
 };
 
